@@ -12,56 +12,57 @@
 #include <config.h>
 #include <stdio.h>
 #include <string.h>
+#include <unistd.h>
 #include <glib-object.h>
 #include <gobject/gmarshal.h>
 #include <bonobo/Bonobo.h>
 #include <bonobo/bonobo-exception.h>
 #include <bonobo/bonobo-main.h>
 #include <bonobo/bonobo-object.h>
+#include <bonobo/bonobo-foreign-object.h>
 #include <bonobo/bonobo-shlib-factory.h>
 #include <bonobo/bonobo-running-context.h>
 #include <bonobo/bonobo-marshal.h>
 #include <bonobo/bonobo-types.h>
 #include <bonobo/bonobo-shutdown.h>
+#include <bonobo/bonobo-debug.h>
+
+/* We need decent ORB cnx. flushing on shutdown to make this work */
+#undef ASYNC_UNREFS
 
 /* Some simple tracking - always on */
 static GMutex *bonobo_total_aggregates_lock = NULL;
 static glong   bonobo_total_aggregates      = 0;
 static glong   bonobo_total_aggregate_refs  = 0;
 
-#ifdef BONOBO_OBJECT_DEBUG
-#	define BONOBO_REF_HOOKS
-#endif
+enum {
+  PROP_0,
+  PROP_POA
+};
 
-/* Define to debug user aggregate merging code */
-#undef BONOBO_AGGREGATE_DEBUG
+/* you may debug by setting BONOBO_DEBUG_FLAGS environment variable to
+   a colon separated list of a subset of {refs,aggregate,lifecycle} */
 
-/* NB. for a quicker debugging experience define this */
-#undef BONOBO_REF_HOOKS 
-
-/* You almost certainly don't want this */
-#undef BONOBO_LIFECYCLE_DEBUG
-
-#ifdef BONOBO_REF_HOOKS
 typedef struct {
 	const char *fn;
 	gboolean    ref;
 	int         line;
 } BonoboDebugRefData;
-#endif
 
 typedef struct {
 	int      ref_count;
 	gboolean immortal;
 	GList   *objs;
-#ifdef BONOBO_REF_HOOKS
+#ifdef G_ENABLE_DEBUG
+	/* the following is required for reference debugging */
 	GList   *refs;
 	int      destroyed;
-#endif
+#endif /* G_ENABLE_DEBUG */
 } BonoboAggregateObject;
 
 struct _BonoboObjectPrivate {
 	BonoboAggregateObject *ao;
+	PortableServer_POA     poa;
 };
 
 enum {
@@ -73,24 +74,9 @@ enum {
 static guint bonobo_object_signals [LAST_SIGNAL];
 static GObjectClass *bonobo_object_parent_class;
 
-#ifdef BONOBO_REF_HOOKS
-
+#ifdef G_ENABLE_DEBUG
 static GHashTable *living_ao_ht = NULL;
-
-static void
-bonobo_debug_print (char *name, char *fmt, ...)
-{
-	va_list args;
-           
-	va_start (args, fmt);
-	
-	printf ("[%06d]:%-15s ", getpid (), name); 
-	vprintf (fmt, args);
-	printf ("\n"); 
-
-	va_end (args);
-}
-#endif /* BONOBO_REF_HOOKS */
+#endif /* G_ENABLE_DEBUG */
 
 /* Do not use this function, it is not what you want; see unref */
 static void
@@ -110,9 +96,10 @@ bonobo_object_destroy (BonoboAggregateObject *ao)
 		} else
 			g_warning ("Serious ref-counting error [%p]", o);
 	}
-#ifdef BONOBO_REF_HOOKS
-	ao->destroyed = TRUE;
-#endif
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) 
+		ao->destroyed = TRUE;
+#endif /* G_ENABLE_DEBUG */
 }
 
 static void
@@ -120,10 +107,13 @@ bonobo_object_corba_deactivate (BonoboObject *object)
 {
 	CORBA_Environment        ev;
 	PortableServer_ObjectId *oid;
+	PortableServer_POA       poa;
 
-#ifdef BONOBO_LIFECYCLE_DEBUG
-	g_warning ("BonoboObject corba deactivate %p", object);
-#endif
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_LIFECYCLE)
+		bonobo_debug_print("deactivate",
+				   "BonoboObject corba deactivate %p", object);
+#endif /* G_ENABLE_DEBUG */
 
 	g_assert (object->priv->ao == NULL);
 
@@ -135,9 +125,9 @@ bonobo_object_corba_deactivate (BonoboObject *object)
 		object->corba_objref = CORBA_OBJECT_NIL;
 	}
 
-	oid = PortableServer_POA_servant_to_id (
-		bonobo_poa(), &object->servant, &ev);
-	PortableServer_POA_deactivate_object (bonobo_poa (), oid, &ev);
+	poa = bonobo_object_get_poa (object);
+	oid = PortableServer_POA_servant_to_id (poa, &object->servant, &ev);
+	PortableServer_POA_deactivate_object (poa, oid, &ev);
 	
 	CORBA_free (oid);
 	CORBA_exception_free (&ev);
@@ -167,16 +157,18 @@ bonobo_object_finalize_internal (BonoboAggregateObject *ao)
 			g_error ("Serious bonobo object corruption");
 		else {
 			g_assert (BONOBO_OBJECT (o)->priv->ao != NULL);
-#ifdef BONOBO_REF_HOOKS
-			g_assert (BONOBO_OBJECT (o)->priv->ao->destroyed);
+#ifdef G_ENABLE_DEBUG
+			if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+				g_assert (BONOBO_OBJECT (o)->priv->ao->destroyed);
 
-			bonobo_debug_print ("finalize", 
-					    "[%p] %-20s corba_objref=[%p]"
-					    " g_ref_count=%d", o,
-					    G_OBJECT_TYPE_NAME (o),
-					    BONOBO_OBJECT (o)->corba_objref,
-					    G_OBJECT (o)->ref_count);
-#endif
+				bonobo_debug_print ("finalize", 
+						    "[%p] %-20s corba_objref=[%p]"
+						    " g_ref_count=%d", o,
+						    G_OBJECT_TYPE_NAME (o),
+						    BONOBO_OBJECT (o)->corba_objref,
+						    G_OBJECT (o)->ref_count);
+			}
+#endif /* G_ENABLE_DEBUG */
 
 			/*
 			 * Disconnect the GObject from the aggregate object
@@ -188,23 +180,28 @@ bonobo_object_finalize_internal (BonoboAggregateObject *ao)
 			 */
 
 			BONOBO_OBJECT (o)->priv->ao = NULL;
-
-			bonobo_object_corba_deactivate (BONOBO_OBJECT (o));
+			if (!g_type_is_a (G_OBJECT_TYPE(o), BONOBO_TYPE_FOREIGN_OBJECT))
+				bonobo_object_corba_deactivate (BONOBO_OBJECT (o));
 
 			g_object_unref (o);
-#ifdef BONOBO_LIFECYCLE_DEBUG
-			g_warning ("Done finalize internal on %p", o);
-#endif
+#ifdef G_ENABLE_DEBUG
+			if(_bonobo_debug_flags & BONOBO_DEBUG_LIFECYCLE)
+				bonobo_debug_print ("finalize",
+						    "Done finalize internal on %p",
+						    o);
+#endif /* G_ENABLE_DEBUG */
 		}
 	}
 
 	g_list_free (ao->objs);
 	ao->objs = NULL;
 
-#ifdef BONOBO_REF_HOOKS
-	for (l = ao->refs; l; l = l->next)
-		g_free (l->data);
-	g_list_free (ao->refs);
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+		for (l = ao->refs; l; l = l->next)
+			g_free (l->data);
+		g_list_free (ao->refs);
+	}
 #endif
 
 	g_free (ao);
@@ -231,9 +228,12 @@ bonobo_object_finalize_servant (PortableServer_Servant servant,
 {
 	BonoboObject *object = bonobo_object(servant);
 
-#ifdef BONOBO_LIFECYCLE_DEBUG
-	g_warning ("BonoboObject Servant finalize %p", object);
-#endif
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_LIFECYCLE)
+		bonobo_debug_print ("finalize",
+				    "BonoboObject Servant finalize %p",
+				    object);
+#endif /* G_ENABLE_DEBUG */
 
 	g_object_unref (G_OBJECT (object));
 }
@@ -241,9 +241,11 @@ bonobo_object_finalize_servant (PortableServer_Servant servant,
 #ifndef bonobo_object_ref
 /**
  * bonobo_object_ref:
- * @object: A BonoboObject you want to ref-count
+ * @obj: A BonoboObject you want to ref-count
  *
  * Increments the reference count for the aggregate BonoboObject.
+ *
+ * Returns: @object
  */
 gpointer
 bonobo_object_ref (gpointer obj)
@@ -256,14 +258,16 @@ bonobo_object_ref (gpointer obj)
 	g_return_val_if_fail (BONOBO_IS_OBJECT (object), object);
 	g_return_val_if_fail (object->priv->ao->ref_count > 0, object);
 
-#ifdef BONOBO_REF_HOOKS
-	bonobo_object_trace_refs (object, "local", 0, TRUE);
-#else
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+		bonobo_object_trace_refs (object, "local", 0, TRUE);
+	}
+	else
+#endif /* G_ENABLE_DEBUG */
 	if (!object->priv->ao->immortal) {
 		object->priv->ao->ref_count++;
 		bonobo_total_aggregate_refs++;
 	}
-#endif
 
 	return object;
 }
@@ -273,40 +277,46 @@ bonobo_object_ref (gpointer obj)
 #ifndef bonobo_object_unref
 /**
  * bonobo_object_unref:
- * @object: A BonoboObject you want to unref.
+ * @obj: A BonoboObject you want to unref.
  *
  * Decrements the reference count for the aggregate BonoboObject.
+ *
+ * Returns: %NULL.
  */
 gpointer
 bonobo_object_unref (gpointer obj)
 {
-#ifndef BONOBO_REF_HOOKS
-	BonoboAggregateObject *ao;
-	BonoboObject *object = obj;
+#ifdef G_ENABLE_DEBUG
+	if(!(_bonobo_debug_flags & BONOBO_DEBUG_REFS)) {
+#endif /* G_ENABLE_DEBUG */
+		BonoboAggregateObject *ao;
+		BonoboObject *object = obj;
 
-	if (!object)
-		return NULL;
+		if (!object)
+			return NULL;
 
-	g_return_val_if_fail (BONOBO_IS_OBJECT (object), NULL);
+		g_return_val_if_fail (BONOBO_IS_OBJECT (object), NULL);
 
-	ao = object->priv->ao;
-	g_return_val_if_fail (ao != NULL, NULL);
-	g_return_val_if_fail (ao->ref_count > 0, NULL);
+		ao = object->priv->ao;
+		g_return_val_if_fail (ao != NULL, NULL);
+		g_return_val_if_fail (ao->ref_count > 0, NULL);
 
-	if (!ao->immortal) {
-		if (ao->ref_count == 1)
-			bonobo_object_destroy (ao);
-
-		ao->ref_count--;
-		bonobo_total_aggregate_refs--;
+		if (!ao->immortal) {
+			if (ao->ref_count == 1)
+				bonobo_object_destroy (ao);
+			
+			ao->ref_count--;
+			bonobo_total_aggregate_refs--;
 		
-		if (ao->ref_count == 0)
-			bonobo_object_finalize_internal (ao);
+			if (ao->ref_count == 0)
+				bonobo_object_finalize_internal (ao);
+		}
+		return NULL;
+#ifdef G_ENABLE_DEBUG
 	}
-	return NULL;
-#else
-	return bonobo_object_trace_refs (obj, "local", 0, FALSE);
-#endif /* BONOBO_REF_HOOKS */
+	else
+		return bonobo_object_trace_refs (obj, "local", 0, FALSE);
+#endif /* G_ENABLE_DEBUG */
 }
 #endif /* bonobo_object_unref */
 
@@ -316,89 +326,90 @@ bonobo_object_trace_refs (gpointer    obj,
 			  int         line,
 			  gboolean    ref)
 {
-#ifdef BONOBO_REF_HOOKS
-	BonoboObject *object = obj;
-	BonoboAggregateObject *ao;
-	BonoboDebugRefData *descr;
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+		BonoboObject *object = obj;
+		BonoboAggregateObject *ao;
+		BonoboDebugRefData *descr;
 
-	if (!object)
-		return NULL;
+		if (!object)
+			return NULL;
 	
-	g_return_val_if_fail (BONOBO_IS_OBJECT (object), ref ? object : NULL);
-	ao = object->priv->ao;
-	g_return_val_if_fail (ao != NULL, ref ? object : NULL);
-
-	descr  = g_new (BonoboDebugRefData, 1);
-	ao->refs = g_list_prepend (ao->refs, descr);
-	descr->fn = fn;
-	descr->ref = ref;
-	descr->line = line;
-
-	if (ref) {
-		g_return_val_if_fail (ao->ref_count > 0, object);
-
-		if (!object->priv->ao->immortal) {
-			object->priv->ao->ref_count++;
-			bonobo_total_aggregate_refs++;
-		}
+		g_return_val_if_fail (BONOBO_IS_OBJECT (object), ref ? object : NULL);
+		ao = object->priv->ao;
+		g_return_val_if_fail (ao != NULL, ref ? object : NULL);
 		
-		bonobo_debug_print ("ref", "[%p]:[%p]:%s to %d at %s:%d", 
-			object, ao,
-		        G_OBJECT_TYPE_NAME (object),
-			ao->ref_count, fn, line);
+		descr  = g_new (BonoboDebugRefData, 1);
+		ao->refs = g_list_prepend (ao->refs, descr);
+		descr->fn = fn;
+		descr->ref = ref;
+		descr->line = line;
 
-		return object;
+		if (ref) {
+			g_return_val_if_fail (ao->ref_count > 0, object);
 
-	} else { /* unref */
-		bonobo_debug_print ("unref", "[%p]:[%p]:%s from %d at %s:%d", 
-			object, ao,
-			G_OBJECT_TYPE_NAME (object),
-			ao->ref_count, fn, line);
+			if (!object->priv->ao->immortal) {
+				object->priv->ao->ref_count++;
+				bonobo_total_aggregate_refs++;
+			}
+		
+			bonobo_debug_print ("ref", "[%p]:[%p]:%s to %d at %s:%d", 
+					    object, ao,
+					    G_OBJECT_TYPE_NAME (object),
+					    ao->ref_count, fn, line);
 
-		g_return_val_if_fail (ao->ref_count > 0, NULL);
+			return object;
+		} else { /* unref */
+			bonobo_debug_print ("unref", "[%p]:[%p]:%s from %d at %s:%d", 
+					    object, ao,
+					    G_OBJECT_TYPE_NAME (object),
+					    ao->ref_count, fn, line);
 
-		if (ao->immortal)
-			bonobo_debug_print ("unusual", "immortal object");
-		else {
-			if (ao->ref_count == 1) {
-				bonobo_object_destroy (ao);
+			g_return_val_if_fail (ao->ref_count > 0, NULL);
+
+			if (ao->immortal)
+				bonobo_debug_print ("unusual", "immortal object");
+			else {
+				if (ao->ref_count == 1) {
+					bonobo_object_destroy (ao);
+					
+					g_return_val_if_fail (ao->ref_count > 0, NULL);
+				}
 				
-				g_return_val_if_fail (ao->ref_count > 0, NULL);
+				/*
+				 * If this blows it is likely some loony used
+				 * g_object_unref somewhere instead of
+				 * bonobo_object_unref, send them my regards.
+				 */
+				g_assert (object->priv->ao == ao);
+				
+				ao->ref_count--;
+				bonobo_total_aggregate_refs--;
+				
+				if (ao->ref_count == 0) {
+					
+					g_assert (g_hash_table_lookup (living_ao_ht, ao) == ao);
+					g_hash_table_remove (living_ao_ht, ao);
+					
+					bonobo_object_finalize_internal (ao);
+					
+				} else if (ao->ref_count < 0) {
+					bonobo_debug_print ("unusual", 
+							    "[%p] already finalized", ao);
+				}
 			}
 			
-			/*
-			 * If this blows it is likely some loony used
-			 * g_object_unref somewhere instead of
-			 * bonobo_object_unref, send them my regards.
-			 */
-			g_assert (object->priv->ao == ao);
-			
-			ao->ref_count--;
-			bonobo_total_aggregate_refs--;
-			
-			if (ao->ref_count == 0) {
-				
-				g_assert (g_hash_table_lookup (living_ao_ht, ao) == ao);
-				g_hash_table_remove (living_ao_ht, ao);
-				
-				bonobo_object_finalize_internal (ao);
-				
-			} else if (ao->ref_count < 0) {
-				bonobo_debug_print ("unusual", 
-						    "[%p] already finalized", ao);
-			}
+			return NULL;
 		}
-
-		return NULL;
 	}
-#else
+	else
+#endif /* G_ENABLE_DEBUG */
 	if (ref)
 		return bonobo_object_ref (obj);
 	else {
 		bonobo_object_unref (obj);
 		return NULL;
 	}
-#endif
 }
 
 static void
@@ -408,11 +419,17 @@ impl_Bonobo_Unknown_ref (PortableServer_Servant servant, CORBA_Environment *ev)
 
 	object = bonobo_object_from_servant (servant);
 
-#if defined(BONOBO_REF_HOOKS) && !defined(bonobo_object_ref)
-	bonobo_object_trace_refs (object, "remote", 0, TRUE);
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+#ifndef bonobo_object_ref
+		bonobo_object_trace_refs (object, "remote", 0, TRUE);
 #else
-	bonobo_object_ref (object);
+		bonobo_object_ref (object);
 #endif
+	}
+	else
+#endif /* G_ENABLE_DEBUG */
+		bonobo_object_ref (object);
 }
 
 void
@@ -467,6 +484,30 @@ bonobo_object_dup_ref (Bonobo_Unknown     object,
 	return ans;
 }
 
+#ifdef ASYNC_UNREFS
+static ORBit_IMethod *
+get_unknown_unref_imethod (void)
+{
+	static ORBit_IMethod *imethod = NULL;
+
+	if (!imethod) {
+		guint i;
+		ORBit_IMethods *methods;
+
+		methods = &Bonobo_Unknown__iinterface.methods;
+
+		for (i = 0; i < methods->_length; i++) {
+			if (!strcmp (methods->_buffer [i].name,
+				     "unref"))
+				imethod = &methods->_buffer [i];
+		}
+		g_assert (imethod != NULL);
+	}
+
+	return imethod;
+}
+#endif
+
 /**
  * bonobo_object_release_unref:
  * @object: a Bonobo_Unknown corba object
@@ -476,6 +517,8 @@ bonobo_object_dup_ref (Bonobo_Unknown     object,
  * it also decrements the ref count on the bonobo object.
  * This is the converse of bonobo_object_dup_ref. We
  * tolerate object == CORBA_OBJECT_NIL silently.
+ *
+ * Returns: %CORBA_OBJECT_NIL.
  **/
 Bonobo_Unknown
 bonobo_object_release_unref (Bonobo_Unknown     object,
@@ -493,6 +536,15 @@ bonobo_object_release_unref (Bonobo_Unknown     object,
 		ev = opt_ev;
 
 	Bonobo_Unknown_unref (object, ev);
+#ifdef ASYNC_UNREFS
+	if (ORBit_small_get_servant (object))
+		Bonobo_Unknown_unref (object, ev);
+	else
+		ORBit_small_invoke_async
+			(object, get_unknown_unref_imethod (),
+			NULL, NULL, NULL, NULL, ev);*/
+#endif
+	
 	CORBA_Object_release (object, ev);
 
 	if (!opt_ev)
@@ -508,11 +560,17 @@ impl_Bonobo_Unknown_unref (PortableServer_Servant servant, CORBA_Environment *ev
 
 	object = bonobo_object_from_servant (servant);
 
-#if defined(BONOBO_REF_HOOKS) && !defined(bonobo_object_unref)
-	bonobo_object_trace_refs (object, "remote", 0, FALSE);
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+#ifndef bonobo_object_unref
+		bonobo_object_trace_refs (object, "remote", 0, FALSE);
 #else
-	bonobo_object_unref (object);
+		bonobo_object_unref (object);
 #endif
+	}
+	else
+#endif /* G_ENABLE_DEBUG */
+		bonobo_object_unref (object);
 }
 
 /**
@@ -569,13 +627,15 @@ impl_Bonobo_Unknown_queryInterface (PortableServer_Servant  servant,
 	local_interface = bonobo_object_query_local_interface (
 		object, repoid);
 
-#ifdef BONOBO_REF_HOOKS
-	bonobo_debug_print ("query-interface", 
-			    "[%p]:[%p]:%s repoid=%s", 
-			    object, object->priv->ao,
-			    G_OBJECT_TYPE_NAME (object),
-			    repoid);
-#endif
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+		bonobo_debug_print ("query-interface", 
+				    "[%p]:[%p]:%s repoid=%s", 
+				    object, object->priv->ao,
+				    G_OBJECT_TYPE_NAME (object),
+				    repoid);
+	}
+#endif /* G_ENABLE_DEBUG */
 
 	if (local_interface == NULL)
 		return CORBA_OBJECT_NIL;
@@ -596,9 +656,13 @@ bonobo_object_finalize_gobject (GObject *gobject)
 {
 	BonoboObject *object = (BonoboObject *) gobject;
 
-#ifdef BONOBO_LIFECYCLE_DEBUG
-	g_warning ("Bonobo Object finalize GObject %p", gobject);
-#endif
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_LIFECYCLE)
+		bonobo_debug_print ("finalize",
+				    "Bonobo Object finalize GObject %p",
+				    gobject);
+#endif /* G_ENABLE_DEBUG */
+
 	if (object->priv->ao != NULL)
 		g_error ("g_object_unreffing a bonobo_object that "
 			 "still has %d refs", object->priv->ao->ref_count);
@@ -614,6 +678,132 @@ bonobo_object_dummy_destroy (BonoboObject *object)
 	/* Just to make chaining possibly cleaner */
 }
 
+static void
+bonobo_object_set_property (GObject         *g_object,
+			    guint            prop_id,
+			    const GValue    *value,
+			    GParamSpec      *pspec)
+{
+	BonoboObject *object = (BonoboObject *) g_object;
+
+	switch (prop_id) {
+	case PROP_POA:
+		object->priv->poa = g_value_get_pointer (value);
+		break;
+	default:
+		break;
+	}
+}
+
+static void
+bonobo_object_get_property (GObject         *g_object,
+			    guint            prop_id,
+			    GValue          *value,
+			    GParamSpec      *pspec)
+{
+	BonoboObject *object = (BonoboObject *) g_object;
+
+	switch (prop_id) {
+	case PROP_POA:
+		g_value_set_pointer (value, object->priv->poa);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
+static void
+do_corba_setup (BonoboObject *object)
+{
+	CORBA_Object obj;
+	CORBA_Environment ev;
+	BonoboObjectClass *xklass;
+	BonoboObjectClass *klass = BONOBO_OBJECT_GET_CLASS (object);
+
+	CORBA_exception_init (&ev);
+
+	/* Setup the servant structure */
+	object->servant._private = NULL;
+	object->servant.vepv     = klass->vepv;
+
+	/* Initialize the servant structure with our POA__init fn */
+	{
+		for (xklass = klass; xklass && !xklass->poa_init_fn;)
+			xklass = g_type_class_peek_parent (xklass);
+		if (!xklass || !xklass->epv_struct_offset) {
+			/*   Also, people using BONOBO_TYPE_FUNC instead of
+			 * BONOBO_TYPE_FUNC_FULL might see this; you need
+			 * to tell it about the CORBA interface you're
+			 * implementing - of course */
+			g_warning ("It looks like you used g_type_unique "
+				   "instead of b_type_unique on type '%s'",
+				   G_OBJECT_CLASS_NAME (klass));
+			return;
+		}
+		xklass->poa_init_fn ((PortableServer_Servant) &object->servant, &ev);
+		if (BONOBO_EX (&ev)) {
+			g_warning ("Exception initializing servant '%s'",
+				   bonobo_exception_get_text (&ev));
+			return;
+		}
+	}
+
+	/*  Instantiate a CORBA_Object reference for the servant
+	 * assumes the bonobo POA supports implicit activation */
+	obj = PortableServer_POA_servant_to_reference (
+		bonobo_object_get_poa (object), &object->servant, &ev);
+
+	if (BONOBO_EX (&ev)) {
+		g_warning ("Exception '%s' getting reference for servant",
+			   bonobo_exception_get_text (&ev));
+		return;
+	}
+
+	object->corba_objref = obj;
+	bonobo_running_context_add_object (obj);
+
+	CORBA_exception_free (&ev);
+}
+
+static GObject *
+bonobo_object_constructor (GType                  type,
+			   guint                  n_construct_properties,
+			   GObjectConstructParam *construct_properties)
+{
+	GObject *g_object;
+	BonoboObject *object;
+
+	g_object = bonobo_object_parent_class->constructor
+		(type, n_construct_properties, construct_properties);
+	if (g_object) {
+		object = (BonoboObject *) g_object;
+
+		/* Though this make look strange, destruction of this object
+		   can only occur when the servant is deactivated by the poa.
+		   The poa maintains its own ref count over method invocations
+		   and delays finalization which happens only after:
+		   bonobo_object_finalize_servant: is invoked */
+		g_object_ref (g_object);
+
+#ifdef G_ENABLE_DEBUG
+		if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+			BonoboAggregateObject *ao = object->priv->ao;
+
+			bonobo_debug_print ("create", "[%p]:[%p]:%s to %d on poa %p", object, ao,
+					    g_type_name (type), ao->ref_count, object->priv->poa);
+
+			g_assert (g_hash_table_lookup (living_ao_ht, ao) == NULL);
+			g_hash_table_insert (living_ao_ht, ao, ao);
+		}
+#endif /* G_ENABLE_DEBUG */
+		if (!g_type_is_a (type, BONOBO_TYPE_FOREIGN_OBJECT))
+			do_corba_setup (object);
+	}
+	
+	return g_object;
+}
+
 /* VOID:CORBA_OBJECT,BOXED */
 static void
 bonobo_marshal_VOID__CORBA_BOXED (GClosure     *closure,
@@ -660,6 +850,12 @@ bonobo_object_class_init (BonoboObjectClass *klass)
 
 	bonobo_object_parent_class = g_type_class_peek_parent (klass);
 
+	object_class->set_property = bonobo_object_set_property;
+	object_class->get_property = bonobo_object_get_property;
+	object_class->constructor  = bonobo_object_constructor;
+	object_class->finalize     = bonobo_object_finalize_gobject;
+	klass->destroy = bonobo_object_dummy_destroy;
+
 	bonobo_object_signals [DESTROY] =
 		g_signal_new ("destroy",
 			      G_TYPE_FROM_CLASS (object_class),
@@ -679,63 +875,11 @@ bonobo_object_class_init (BonoboObjectClass *klass)
 			      BONOBO_TYPE_STATIC_CORBA_OBJECT,
 			      BONOBO_TYPE_STATIC_CORBA_EXCEPTION);
 
-	klass->destroy = bonobo_object_dummy_destroy;
-
-	object_class->finalize = bonobo_object_finalize_gobject;
-}
-
-
-static void
-do_corba_setup (BonoboObject      *object,
-		BonoboObjectClass *klass)
-{
-	CORBA_Object obj;
-	CORBA_Environment ev;
-	BonoboObjectClass *xklass;
-
-	CORBA_exception_init (&ev);
-
-	/* Setup the servant structure */
-	object->servant._private = NULL;
-	object->servant.vepv     = klass->vepv;
-
-	/* Initialize the servant structure with our POA__init fn */
-	{
-		for (xklass = klass; xklass && !xklass->poa_init_fn;)
-			xklass = g_type_class_peek_parent (xklass);
-		if (!xklass || !xklass->epv_struct_offset) {
-			/*   Also, people using BONOBO_TYPE_FUNC instead of
-			 * BONOBO_TYPE_FUNC_FULL might see this; you need
-			 * to tell it about the CORBA interface you're
-			 * implementing - of course */
-			g_warning ("It looks like you used g_type_unique "
-				   "instead of b_type_unique on type '%s'",
-				   G_OBJECT_CLASS_NAME (klass));
-			return;
-		}
-		xklass->poa_init_fn ((PortableServer_Servant) &object->servant, &ev);
-		if (BONOBO_EX (&ev)) {
-			g_warning ("Exception initializing servant '%s'",
-				   bonobo_exception_get_text (&ev));
-			return;
-		}
-	}
-
-	/*  Instantiate a CORBA_Object reference for the servant
-	 * assumes the bonobo POA supports implicit activation */
-	obj = PortableServer_POA_servant_to_reference (
-		bonobo_poa (), &object->servant, &ev);
-
-	if (BONOBO_EX (&ev)) {
-		g_warning ("Exception '%s' getting reference for servant",
-			   bonobo_exception_get_text (&ev));
-		return;
-	}
-
-	object->corba_objref = obj;
-	bonobo_running_context_add_object (obj);
-
-	CORBA_exception_free (&ev);
+	g_object_class_install_property
+		(object_class, PROP_POA,
+		 g_param_spec_pointer
+			("poa", _("POA"), _("Custom CORBA POA"),
+			 G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 }
 
 static void
@@ -745,11 +889,15 @@ bonobo_object_instance_init (GObject    *g_object,
 	BonoboObject *object = BONOBO_OBJECT (g_object);
 	BonoboAggregateObject *ao;
 
-#ifdef BONOBO_OBJECT_DEBUG
-	g_warning ("bonobo_object_instance init '%s' '%s' -> %p",
-		   G_OBJECT_TYPE_NAME (g_object),
-		   G_OBJECT_CLASS_NAME (klass), object);
-#endif
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_OBJECT) {
+		bonobo_debug_print ("object",
+				    "bonobo_object_instance init '%s' '%s' -> %p",
+				    G_OBJECT_TYPE_NAME (g_object),
+				    G_OBJECT_CLASS_NAME (klass), object);
+	}
+#endif /* G_ENABLE_DEBUG */
+
 	/* Some simple debugging - count aggregate allocate */
 	LINC_MUTEX_LOCK   (bonobo_total_aggregates_lock);
 	bonobo_total_aggregates++;
@@ -764,30 +912,11 @@ bonobo_object_instance_init (GObject    *g_object,
 	/* Setup Private fields */
 	object->priv = g_new (BonoboObjectPrivate, 1);
 	object->priv->ao = ao;
+	object->priv->poa = NULL;
 
 	/* Setup signatures */
 	object->object_signature  = BONOBO_OBJECT_SIGNATURE;
 	object->servant_signature = BONOBO_SERVANT_SIGNATURE;
-
-	/* Though this make look strange, destruction of this object
-	   can only occur when the servant is deactivated by the poa.
-	   The poa maintains its own ref count over method invocations
-	   and delays finalization which happens only after:
-	   bonobo_object_finalize_servant: is invoked */
-	g_object_ref (g_object);
-
-#ifdef BONOBO_REF_HOOKS
-	{
-		bonobo_debug_print ("create", "[%p]:[%p]:%s to %d", object, ao,
-				    g_type_name (G_TYPE_FROM_CLASS (klass)),
-				    ao->ref_count);
-
-		g_assert (g_hash_table_lookup (living_ao_ht, ao) == NULL);
-		g_hash_table_insert (living_ao_ht, ao, ao);
-	}
-#endif
-
-	do_corba_setup (object, BONOBO_OBJECT_CLASS (klass));
 }
 
 /**
@@ -813,18 +942,19 @@ bonobo_object_get_type (void)
 			(GInstanceInitFunc) bonobo_object_instance_init
 		};
 		
+#ifdef G_ENABLE_DEBUG
+		bonobo_debug_init();
+		if(_bonobo_debug_flags & BONOBO_DEBUG_REFS)
+			living_ao_ht = g_hash_table_new (NULL, NULL);
+#endif /* G_ENABLE_DEBUG */
 		type = g_type_register_static (G_TYPE_OBJECT, "BonoboObject",
 					       &info, 0);
-
-#ifdef BONOBO_REF_HOOKS
-		living_ao_ht = g_hash_table_new (NULL, NULL);
-#endif
 	}
 
 	return type;
 }
 
-#ifdef BONOBO_REF_HOOKS
+#ifdef G_ENABLE_DEBUG
 static void
 bonobo_ao_debug_foreach (gpointer key, gpointer value, gpointer user_data)
 {
@@ -860,27 +990,29 @@ bonobo_ao_debug_foreach (gpointer key, gpointer value, gpointer user_data)
 				    descr->fn, descr->line);
 	}
 }
-#endif
+#endif /* G_ENABLE_DEBUG */
 
 int
 bonobo_object_shutdown (void)
 {
-#ifdef BONOBO_REF_HOOKS
-	
-	bonobo_debug_print ("shutdown-start", 
-		"-------------------------------------------------");
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {	
+		bonobo_debug_print ("shutdown-start", 
+				    "-------------------------------------------------");
 
-	if (living_ao_ht)
-		g_hash_table_foreach (living_ao_ht,
-				      bonobo_ao_debug_foreach, NULL);
+		if (living_ao_ht)
+			g_hash_table_foreach (living_ao_ht,
+					      bonobo_ao_debug_foreach, NULL);
+		
+		bonobo_debug_print ("living-objects",
+				    "living bonobo objects count = %d",
+				    g_hash_table_size (living_ao_ht));
+		
+		bonobo_debug_print ("shutdown-end", 
+				    "-------------------------------------------------");
+	}
+#endif /* G_ENABLE_DEBUG */
 
-	bonobo_debug_print ("living-objects",
-			    "living bonobo objects count = %d",
-			    g_hash_table_size (living_ao_ht));
-
-	bonobo_debug_print ("shutdown-end", 
-		"-------------------------------------------------");
-#endif
 	if (bonobo_total_aggregates > 0) {
 		g_warning ("Leaked a total of %ld refs to %ld bonobo object(s)",
 			   bonobo_total_aggregate_refs,
@@ -935,50 +1067,54 @@ bonobo_object_add_interface (BonoboObject *object, BonoboObject *newobj)
        ao->ref_count = ao->ref_count + oldao->ref_count - 1;
        bonobo_total_aggregate_refs--;
 
-#ifdef BONOBO_REF_HOOKS		
-       bonobo_debug_print ("add_interface", 
-			   "[%p]:[%p]:%s to [%p]:[%p]:%s ref_count=%d", 
-			   object, object->priv->ao,
-			   G_OBJECT_TYPE_NAME (object),
-			   newobj, newobj->priv->ao,
-			   G_OBJECT_TYPE_NAME (newobj),
-			   ao->ref_count);
-#endif
+#ifdef G_ENABLE_DEBUG
+       if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+	       bonobo_debug_print ("add_interface", 
+				   "[%p]:[%p]:%s to [%p]:[%p]:%s ref_count=%d", 
+				   object, object->priv->ao,
+				   G_OBJECT_TYPE_NAME (object),
+				   newobj, newobj->priv->ao,
+				   G_OBJECT_TYPE_NAME (newobj),
+				   ao->ref_count);
+       }
+#endif /* G_ENABLE_DEBUG */
 
        /* Merge the two AggregateObject lists */
        for (l = oldao->objs; l; l = l->next) {
 	       BonoboObject *new_if = l->data;
 
-#ifdef BONOBO_AGGREGATE_DEBUG
-	       GList *i;
-	       CORBA_Environment ev;
-	       CORBA_char *new_id;
+#ifdef G_ENABLE_DEBUG
+	       if(_bonobo_debug_flags & BONOBO_DEBUG_AGGREGATE) {
+		       GList *i;
+		       CORBA_Environment ev;
+		       CORBA_char *new_id;
+		       
+		       CORBA_exception_init (&ev);
 
-	       CORBA_exception_init (&ev);
+		       new_id = ORBit_small_get_type_id (new_if->corba_objref, &ev);
 
-	       new_id = ORBit_small_get_type_id (new_if->corba_objref, &ev);
+		       for (i = ao->objs; i; i = i->next) {
+			       BonoboObject *old_if = i->data;
 
-	       for (i = ao->objs; i; i = i->next) {
-		       BonoboObject *old_if = i->data;
-
-		       if (old_if == new_if)
-			       g_error ("attempting to merge identical "
-					"interfaces [%p]", new_if);
-		       else {
-			       CORBA_char *old_id;
-
-			       old_id = ORBit_small_get_type_id (old_if->corba_objref, &ev);
-			       
-			       if (!strcmp (new_id, old_id))
-				       g_error ("Aggregating two BonoboObject that implement "
-						"the same interface '%s' [%p]", new_id, new_if);
-			       CORBA_free (old_id);
+			       if (old_if == new_if)
+				       g_error ("attempting to merge identical "
+						"interfaces [%p]", new_if);
+			       else {
+				       CORBA_char *old_id;
+				       
+				       old_id = ORBit_small_get_type_id (old_if->corba_objref, &ev);
+				       
+				       if (!strcmp (new_id, old_id))
+					       g_error ("Aggregating two BonoboObject that implement "
+							"the same interface '%s' [%p]", new_id, new_if);
+				       CORBA_free (old_id);
+			       }
 		       }
+		       
+		       CORBA_free (new_id);
+		       CORBA_exception_free (&ev);
 	       }
-
-	       CORBA_free (new_id);
-	       CORBA_exception_free (&ev);
-#endif
+#endif /* G_ENABLE_DEBUG */
 
 	       ao->objs = g_list_prepend (ao->objs, new_if);
 	       new_if->priv->ao = ao;
@@ -986,11 +1122,13 @@ bonobo_object_add_interface (BonoboObject *object, BonoboObject *newobj)
 
        g_assert (newobj->priv->ao == ao);
 
-#ifdef BONOBO_REF_HOOKS
-       g_assert (g_hash_table_lookup (living_ao_ht, oldao) == oldao);
-       g_hash_table_remove (living_ao_ht, oldao);
-       ao->refs = g_list_concat (ao->refs, oldao->refs);
-#endif
+#ifdef G_ENABLE_DEBUG
+       if(_bonobo_debug_flags & BONOBO_DEBUG_REFS) {
+	       g_assert (g_hash_table_lookup (living_ao_ht, oldao) == oldao);
+	       g_hash_table_remove (living_ao_ht, oldao);
+	       ao->refs = g_list_concat (ao->refs, oldao->refs);
+       }
+#endif /* G_ENABLE_DEBUG */
 
        g_list_free (oldao->objs);
        g_free (oldao);
@@ -1321,10 +1459,12 @@ bonobo_type_setup (GType             type,
 		return FALSE;
 	}
 
-#ifdef BONOBO_OBJECT_DEBUG
-	g_warning ("We are at depth %d with type '%s'",
-		   depth, g_type_name (type));
-#endif
+#ifdef G_ENABLE_DEBUG
+	if(_bonobo_debug_flags & BONOBO_DEBUG_OBJECT) {
+		bonobo_debug_print ("object", "We are at depth %d with type '%s'",
+				    depth, g_type_name (type));
+	}
+#endif /* G_ENABLE_DEBUG */
 
 	/* Setup the Unknown epv */
 	bonobo_object_epv_init (&klass->epv);
@@ -1370,6 +1510,7 @@ bonobo_type_setup (GType             type,
  * commences at, or 0 if we are inheriting a plain GObject
  * from a BonoboObject, adding no new CORBA interfaces
  * @info: the standard GTypeInfo.
+ * @type_name: the name of the type being registered.
  * 
  * This function is the main entry point for deriving bonobo
  * server interfaces.
@@ -1405,3 +1546,62 @@ bonobo_type_unique (GType             parent_type,
 	else
 		return 0;
 }
+
+/**
+ * bonobo_object_query_remote:
+ * @unknown: an unknown object ref ( or NIL )
+ * @repo_id: the interface to query for
+ * @opt_ev: an optional exception environment
+ * 
+ * A helper wrapper for query interface
+ * 
+ * Return value: the interface or CORBA_OBJECT_NIL
+ **/
+Bonobo_Unknown
+bonobo_object_query_remote (Bonobo_Unknown     unknown,
+			    const char        *repo_id,
+			    CORBA_Environment *opt_ev)
+{
+	Bonobo_Unknown new_if;
+	CORBA_Environment *ev, temp_ev;
+       
+	if (unknown == CORBA_OBJECT_NIL)
+		return CORBA_OBJECT_NIL;
+
+	if (!opt_ev) {
+		CORBA_exception_init (&temp_ev);
+		ev = &temp_ev;
+	} else
+		ev = opt_ev;
+
+	new_if = Bonobo_Unknown_queryInterface (
+		unknown, repo_id, ev);
+
+	if (BONOBO_EX (ev))
+		new_if = CORBA_OBJECT_NIL;
+
+	if (!opt_ev)
+		CORBA_exception_free (ev);
+
+	return new_if;
+}
+
+/**
+ * bonobo_object_get_poa:
+ * @object: the object associated with an interface
+ * 
+ * Gets the POA associated with this part of the
+ * BonoboObject aggregate it is possible to have
+ * different POAs per interface.
+ * 
+ * Return value: the poa, never NIL.
+ **/
+PortableServer_POA
+bonobo_object_get_poa (BonoboObject *object)
+{
+	g_return_val_if_fail (object != CORBA_OBJECT_NIL, bonobo_poa ());
+	if (object->priv->poa)
+		return object->priv->poa;
+	else
+		return bonobo_poa ();
+}
